from __future__ import annotations

from typing import Any
from typing import ClassVar
from typing import Pattern
from typing import TypeVar

from phantom.base import Phantom

P = TypeVar("P", bound=Pattern)


class Match(str, Phantom):
    __pattern__: ClassVar[Pattern]

    def __init_subclass__(cls, *, pattern: Pattern) -> None:
        super().__init_subclass__()
        cls.__pattern__ = pattern

    @classmethod
    def __instancecheck__(cls, instance: Any) -> bool:
        return isinstance(instance, str) and cls.__pattern__.match(instance) is not None



from typing import Any
from typing import Type
from typing import TypeVar

from .base import Phantom


T = TypeVar("T", bound=float)


class OpenRange(Phantom):
    __min__: float
    __max__: float
    __type__: Type[float]

    def __init_subclass__(
        cls, *, min: float = float("-inf"), max: float = float("inf")
    ):
        super().__init_subclass__()
        cls.__min__ = min
        cls.__max__ = max
        if not issubclass(cls.__bases__[0], (int, float)):
            raise TypeError(
                "The first base of subclasses of OpenRange must be a subclass "
                "of either int or float."
            )
        cls.__type__ = cls.__bases__[0]

    @classmethod
    def __instancecheck__(cls, instance: Any) -> bool:
        return (
            isinstance(instance, cls.__type__)
            and cls.__min__ <= instance <= cls.__max__
        )


class Natural(int, OpenRange, min=0):
    ...


class NegativeInt(int, OpenRange, max=0):
    ...


class Portion(float, OpenRange, min=0, max=1):
    ...



import datetime
from typing import Any

from .base import Phantom


def is_tz_aware(dt: datetime.datetime) -> bool:
    # https://docs.python.org/3/library/datetime.html#determining-if-an-object-is-aware-or-naive
    return dt.tzinfo is not None and dt.tzinfo.utcoffset(dt) is not None


class TZAware(datetime.datetime, Phantom):
    @classmethod
    def __instancecheck__(cls, instance: Any) -> bool:
        return isinstance(instance, datetime.datetime) and is_tz_aware(instance)


class TZNaive(datetime.datetime, Phantom):
    @classmethod
    def __instancecheck__(cls, instance: Any) -> bool:
        return isinstance(instance, datetime.datetime) and not is_tz_aware(instance)



import abc
from typing import Any
from typing import Type
from typing import TypeVar


class PhantomMeta(abc.ABCMeta):
    """
    Metaclass that defers __instancecheck__ to derived classes and prevents
    actual instance creation.
    """

    def __instancecheck__(self, instance: Any) -> bool:
        if not issubclass(self, Phantom):
            return False
        return self.__instancecheck__(instance)

    def __call__(cls, instance):
        return cls.from_instance(instance)  # type: ignore[attr-defined]


Derived = TypeVar("Derived")


class Phantom(metaclass=PhantomMeta):
    @classmethod
    def from_instance(cls: Type[Derived], instance: Any) -> Derived:
        if not isinstance(instance, cls):
            raise TypeError(
                f"Can't create phantom type {cls.__qualname__} from {instance!r}"
            )
        return instance

    @classmethod
    @abc.abstractmethod
    def __instancecheck__(cls, instance: Any) -> bool:
        ...



import re

import pytest

from phantom.re import Match


class A(Match, pattern=re.compile(r"^abc$")):
    ...


class TestMatch:
    def test_non_matching_string_is_not_instance(self):
        assert not isinstance("a", A)

    def test_matching_string_is_instance(self):
        assert isinstance("abc", A)

    def test_instantiation_raises_for_non_matching_string(self):
        with pytest.raises(TypeError):
            A.from_instance("b")

    def test_instantiation_returns_instance(self):
        s = "abc"
        assert s is A.from_instance(s)



# This is the closest I could find to documentation of _ProtocolMeta ...
# https://github.com/python/cpython/commit/74d7f76e2c953fbfdb7ce01b7319d91d471cc5ef
from typing import _ProtocolMeta  # type: ignore[attr-defined]
from typing import Any
from typing import Final
from typing import Generic
from typing import Iterable
from typing import MutableMapping
from typing import MutableSequence
from typing import MutableSet
from typing import Optional
from typing import Protocol
from typing import runtime_checkable
from typing import Sized
from typing import Type
from typing import TypeVar

from .base import Phantom
from .base import PhantomMeta


__all__ = (
    "PhantomSized",
    "NonEmpty",
    "Empty",
)


mutable: Final = (MutableSequence, MutableSet, MutableMapping)
T = TypeVar("T", bound=Any, covariant=True)


@runtime_checkable
class SizedIterable(Sized, Iterable[T], Protocol[T]):
    ...


class SizedIterablePhantomMeta(PhantomMeta, _ProtocolMeta):
    ...


class PhantomSized(
    SizedIterable[T], Phantom, Generic[T], metaclass=SizedIterablePhantomMeta
):
    __min__: int
    __max__: float

    def __init_subclass__(
        cls, *, min: Optional[int] = None, max: Optional[float] = None
    ):
        super().__init_subclass__()
        # Resolve __min__ and __max__ in the order: class argument, inherited
        # value, default.
        cls.__min__ = getattr(cls, "__min__", 0) if min is None else min
        cls.__max__ = getattr(cls, "__max__", float("inf")) if max is None else max

    @classmethod
    def __instancecheck__(cls, instance: Any) -> bool:
        return (
            isinstance(instance, SizedIterable)
            and not isinstance(instance, mutable)
            and cls.__min__ <= len(instance) <= cls.__max__
        )

    def __class_getitem__(cls: Type[T], item: Type[SizedIterable]) -> Type[T]:
        return cls


class NonEmpty(PhantomSized[T], Generic[T], min=1):
    ...


class Empty(PhantomSized[T], Generic[T], min=0, max=0):
    ...
